-------------------------------------------------------------------------------- 
-- UPC - EETAC - EEL - CSD - DIGSYS       
--------------------------------------------------------------------------------
-- P6. Example tutorial. Classroom lightning system 
-- Plan: Synchronous FSM (Plan C1: several processes in the same VHDL file)
-- This file is the translation of the ideas exposed in the planning section of 
-- https://digsys.upc.edu/csd/P06/lab06/lab6.html
-- Adaptation of the FSM at: 
-- https://digsys.upc.edu/csd/P06/lab06/img_FSM.jpg
--------------------------------------------------------------------------------

LIBRARY ieee;     
USE  IEEE.STD_LOGIC_1164.all;
USE  IEEE.STD_LOGIC_UNSIGNED.all;
			
ENTITY Light_Control IS
      Port (	CLK 	: IN	STD_LOGIC;
				B		: IN	STD_LOGIC;
				CD   	: IN 	STD_LOGIC;
				Z 		: OUT	STD_LOGIC;
				-- Extra output to monitor the current state 
				CS_out	: OUT	STD_LOGIC_VECTOR(2 downto 0)
	);
END Light_Control;

ARCHITECTURE FSM_like OF Light_Control IS

	TYPE State_type IS (Idle, Click_1_detected, Button_released_1,
					Click_2_detected, Button_released_2, Click_3_detected);  

	SIGNAL current_state, next_state	: State_type ; 	
------------------------------------------------------------------------------
-- Synthesis tools have options to choose how to encode FSM_: 
-- "one_hot", "sequential" (this is radix-2)", "Johnson", "gray" and "auto".

BEGIN

	-------- The state register that contains the memory cells
state_register: PROCESS (CD, CLK, next_state)
	BEGIN
						
	IF CD = '1' THEN 	-- reset state where to go if a CD signal is asserted 					
		current_state <= Idle; 
	ELSIF (CLK'EVENT and CLK = '1') THEN -- This is the description of D_FF
		current_state <= next_state; 
	END IF;

END PROCESS state_register;

	-------- CC1: Combinational system for calculating next state
	-------- The truth table that contains all the state transitions (arrows)

CC_1: PROCESS (current_state, B)
	BEGIN
	   
		CASE current_state IS
				
			WHEN Idle =>
			   IF (B = '0') THEN
					next_state <= Idle;	
				ELSE    
					next_state <= Click_1_detected; 
				END IF;
				
			WHEN Click_1_detected =>  
				IF (B = '1') THEN
					next_state <= Click_1_detected; 
				ELSE    
					next_state <= Button_released_1 ;
				END IF;
				
			WHEN Button_released_1 => 
				IF (B = '0') THEN
					next_state <= Button_released_1; 
				ELSE    
					next_state <= Click_2_detected ;
				END IF;
				
			WHEN Click_2_detected => 
				IF (B = '1') THEN
					next_state <= Click_2_detected; 
				ELSE    
					next_state <= Button_released_2 ;
				END IF;
				
			WHEN Button_released_2 => 
			   IF (B = '0') THEN
			      next_state <= Button_released_2; 
				ELSE    
					next_state <= Click_3_detected ;
				END IF;

			WHEN Click_3_detected => 
			   IF (B = '1') THEN
			      next_state <= Click_3_detected; 
				ELSE    
					next_state <= Idle ; -- Go bakc to the initial state
				END IF;
						
		END CASE ;		

END PROCESS CC_1;

	-------- CS_2: combinational system for calculating the outputs 
CC_2: PROCESS (current_state)
	BEGIN
		CASE current_state IS
			WHEN Idle => 
				Z <= '0';

			WHEN Click_1_detected => 
				Z <= '0'; 	

			WHEN Button_released_1 => 
				Z <= '0'; 	

			WHEN Click_2_detected => 
				Z <= '1'; -- Now the light has to be turned ON	

			WHEN Button_released_2 => 
				Z <= '1'; -- keep the light ON 

			WHEN Click_3_detected => 
				Z <= '0'; 	-- Switch OFF the light 
		END CASE ;		

END PROCESS CC_2;

-- Place here other logic if necessary 
---> current state outputs for debugging purposes


--Current_State_decoding:  combinational system for calculating the outputs 
-- Plan B truth table description. Any kind of encoding can be adapted to BCD
-- for representing 7-segment numbers
------------------------------------------------------------------------------
CC_3: PROCESS (current_state)
	BEGIN
		CASE current_state IS
			WHEN Idle => 
				CS_out <= "000";

			WHEN Click_1_detected => 
				CS_out <= "001";

			WHEN Button_released_1 => 
				CS_out <= "010";

			WHEN Click_2_detected => 
				CS_out <= "011";

			WHEN Button_released_2 => 
				CS_out <= "100";

			WHEN Click_3_detected => 
				CS_out <= "101";
		END CASE ;		

END PROCESS CC_3;

END FSM_like;